/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/*
 * convert hexadecimal value into decimal value
 **************************************************/
uint64_t get_hex(char *buf){
  uint64_t sum  = 0;

  int num = strlen(buf);
  if(num>16){
    printf("Error! hex() can treat within 16digit numbers.\n");
    exit(1);
  }
  
  for(int i=0; i<num; i++){
    uint64_t temp = 0;
    char t = *((char *)(buf+(num-i-1)));
    switch(t){
    case '0': temp = 0; break;
    case '1': temp = 1; break;
    case '2': temp = 2; break;
    case '3': temp = 3; break;
    case '4': temp = 4; break;
    case '5': temp = 5; break;
    case '6': temp = 6; break;
    case '7': temp = 7; break;
    case '8': temp = 8; break;
    case '9': temp = 9; break;
    case 'a': temp = 10;break;
    case 'b': temp = 11;break;
    case 'c': temp = 12;break;
    case 'd': temp = 13;break;
    case 'e': temp = 14;break;
    case 'f': temp = 15;break;
    }
    sum += temp * (uint64_t)(1llu << 4*i);
  }
  return sum;
}

/************************************************/
void print_register_name(int i){
  switch(i){
  case 0:   printf("v0    "); break;
  case 1:   printf("t0    "); break;
  case 2:   printf("t1    "); break;
  case 3:   printf("t2    "); break;
  case 4:   printf("t3    "); break;
  case 5:   printf("t4    "); break;
  case 6:   printf("t5    "); break;
  case 7:   printf("t6    "); break;
  case 8:   printf("t7    "); break;
  case 9:   printf("s0    "); break;
  case 10:  printf("s1    "); break;
  case 11:  printf("s2    "); break;
  case 12:  printf("s3    "); break;
  case 13:  printf("s4    "); break;
  case 14:  printf("s5    "); break;
  case 15:  printf("fp    "); break;
  case 16:  printf("a0    "); break;
  case 17:  printf("a1    "); break;
  case 18:  printf("a2    "); break;
  case 19:  printf("a3    "); break;
  case 20:  printf("a4    "); break;
  case 21:  printf("a5    "); break;
  case 22:  printf("t8    "); break;
  case 23:  printf("t9    "); break;
  case 24:  printf("t10   "); break;
  case 25:  printf("t11   "); break;
  case 26:  printf("ra    "); break;
  case 27:  printf("t12   "); break;
  case 28:  printf("at    "); break;
  case 29:  printf("gp    "); break;
  case 30:  printf("sp    "); break;
  case 31:  printf("zero  "); break;
  case 32:  printf("pc    "); break;
  case 33:  printf("vfp   "); break;
  }
}

/************************************************/
void print_status(ADDR_TYPE pc, char *buf, data_t *reg, uint64_t tc){
  for(int i=0; i<32; i++){
    print_register_name(i);
    printf("r%-2d      0x%-6qx %-8qd\n", i, reg[i].ld(), reg[i].ld());
  }
  
  printf("fpcr           0x%-6qx %-8qd\n", 0ull, 0ull);
  printf("pc             0x%-6qx %-8qd\n", pc,   pc  );
  printf("vfp            0x%-6qx %-8qd\n", 0ull, 0ull);
}


/************************************************/
debug::debug(architecture_state *as_t, memory_system *mem_t,
	     system_config *ms_t, evaluation_result *e_t){
  as  =  as_t;
  mem = mem_t;
  ms  =  ms_t;
  e   =   e_t;
}

/************************************************/
void debug::debugmode(){
  int i,j;
  char t_buf[256];
  char t_temp[256];
  static int stop = 1;

  char *buf;
  char *temp;
  
  static unsigned long long stop_count = 0;
  static unsigned long long stop_address = 0xFFFFFFFFFFFFFFFFull; 
  unsigned long long address;

  buf = t_buf;
  temp = t_temp;

  if(stop_count==e->retired_inst){
    stop = 1;
  }

  if(stop_address==as->pc.ld()){
    stop = 1;
  }

  if(stop==0) return;
  
  if(stop==1){
    data_t d;
    INST_TYPE ir;
    mem->ld_nbyte(4, &as->pc, &d);
    ir = d.ld() & 0xffffffff;

    data_t ir_t;
    mem->ld_nbyte(4, &as->pc, &ir_t);
    printf("TC:%qd  pc:%qx  ir:%08qx\n",
	   e->retired_inst, as->pc.ld(), ir_t.ld());
    
    print_status(as->pc.ld(), NULL, as->r, e->retired_inst);
    
    printf("(db)");
    fgets(buf,256, stdin);

    while(*buf != (char)NULL){
      
      if(!strncmp(buf,"quit",4)) exit(1);
      if(!strncmp(buf,"h",1)){ /* help */
	printf("  command\n");
	printf("  c followd by number:  stop specified tc.\n");
	printf("  s followd by address: stop specified address.\n");
	printf("  x followd by address: examine specified address.\n");
	printf("  quit: quit simulate.\n");
      }

      if(*(buf)=='\n') break;

      if(*(buf)=='x'){
	buf++;
	address = get_hex(buf);
	address = address >> 4; /* I don't know why? */

	for(j=0; j<40; j++){
	  data_t tmp;
	  printf("0x%x%x: ",(int)(address >> 32),(int)(address & MASK32));
	  for(i=0; i<2; i++){
	    data_t adr;
	    adr.st(address);
	    mem->ld_nbyte(8, &adr, &tmp);
	    printf("0x%08x%08x ", (int)(tmp.ld() >> 32), 
		   (int)(tmp.ld() & MASK32));
	    address +=8;
	  }
	  printf("\n");
	}
      }
      
      if(*(buf)=='c'){ /* stop by total count */
	buf++;
	stop_count = atol(buf); /* for int use atoi */
	printf("stop_tc = %qd\n", stop_count);
	stop=0;
	break;
      }

      if(*(buf)=='s'){ /* stop by address */
	buf++;
	stop_address  = get_hex(buf);
	stop_address  = stop_address >> 4;
	printf("stop_address = 0x%qx\n", stop_address);
	stop=0;
	break;
      }
      
      printf("\n(db)");
      fgets(buf,256,stdin);
    }
  }
}
/************************************************/
