/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/************************************************/
void usage(){
  char UsageMessage[] =  "\
Usage: SimAlpha [-option] command\n\
 -d: debug mode\n\
 -c: print systemcall trace\n\
 -v[num]: verbose mode, prints '.'\
by VERBOSE_INTERVAL\n\
 -f[num]: set fd_adjust to verify simulation\n\
 -e[num][m]: end simulation at num code run. \
m means million\n\
 -i[file_name]: file_name as STDIN input file\n\
";
  printf("%s", UsageMessage); exit(0);
}

/************************************************/
void option_from_file(char **opt, char *name){
  char *buf;
  for(int i=0; i<OPTION_MAX; i++) opt[i]=NULL;
  FILE *fp = fopen(name,"r");
  if(fp!=NULL){
    printf("*** Use %s\n", name);
    for(int i=0; !feof(fp);){
      buf = (char *)malloc(256);
      fscanf(fp, "%s", buf);
      if(buf[0]=='-') opt[i++] = buf;
    }
  }
}

/* C interface                                  */
/************************************************/
int simple_cpu(int job, uint64_t *p){
  static simple_chip *chip;
	
  switch(job){
  case 1:{
    printf("%s\n", VER); 
    static char *opt[OPTION_MAX];
    option_from_file(opt, "SimAlpha_Option.txt");
    chip = new simple_chip(DEF_FILE_NAME, opt);
  }
  break;
  case 2:{
    *(p++) = chip->as->pc.ld();
    int running = chip->step();
    
    for(int i=0; i<31; i++) 
      *(p++)=chip->as->r[i].ld();
    for(int i=0; i<31; i++) 
      *(p++)=chip->as->f[i].ld();
    
    if(running==0) delete chip;
    break;
  }
  default: 
    printf("Warning: %d siple_cpu.\n", job);
  }
  return 0;
}

/* map_s is used in LDS & ITOFS instruction     */
/* input of 8bit and output of 11bit            */
/************************************************/
uint64_t map_s(unsigned int int_tmp){
  unsigned int i8 = (int_tmp >> 23) & 0xFF;

  /** Next line should be valid, 
      SimpleScalar has a bug ? **/
  // if(i8==0xFF)  return 0x00000000000007FFull;
  if(i8==0x00)  return 0x0000000000000000ull;
  
  if(i8 & 0x80) return 0x400ull | (i8 & 0x7F);
  else          return 0x380ull | (i8 & 0x7F);

  printf("Warning In load_store.c, map_s(). \n");
}

/* If IR is not the cmov, return 0.             */
/************************************************/
INST_TYPE cmov_ir_create(INST_TYPE ir){
  int Op     = (ir>>26) & 0x3F;
  int func7  = (ir>>5)  & 0x7f;
  int func11 = (ir>>5)  & 0x7FF;

  switch(Op){
  case 0x11:
    switch(func7){
    case 0x14: 
    case 0x16: 
    case 0x24: 
    case 0x26: 
    case 0x44: 
    case 0x46: 
    case 0x64: 
    case 0x66: 
      {
	INST_TYPE ret;
	int Op  = (ir>>26) &  0x3F;
	int Lit = (ir>>13) &  0xFF;
	int RC  = (ir    ) &  0x1F;
	int Fnc = 0x6e; /** This is assigned by Kenji KISE. **/
	ret = (Op << 26) | (RC<<21) | (Lit<<13) | (ir & BIT12)
	  | (Fnc<<5) | RC;
	return ret;
      }
    }
    
  case 0x17:
    switch(func11){
    case 0x02a:
    case 0x02b:
    case 0x02c:
    case 0x02d:
    case 0x02e:
    case 0x02f:
      {
	INST_TYPE ret;
	int Op  = (ir>>26) &  0x3F;
	int RB  = (ir>>16) &  0x1F;
	int RC  = (ir    ) &  0x1F;
	int Fnc = 0x333; /** This is assigned by Kenji KISE. **/
	ret = (Op << 26) | (RC<<21) | (RB<<16) | (Fnc<<5) | RC;
	return ret;
      }
    }
  }
  return 0;
}

/************************************************/
int data_t::init(uint64_t d){
  value = d;
  cmov  = 0;
  return 0;
}

/************************************************/
uint64_t data_t::ld(){
  return value;
}

/************************************************/
int data_t::st(uint64_t d){
  value = d;
  return 0;
}

/*
 * Initialize the registers, data file format.
 * Begin by "\@" to indicate the valid data.
 * /@reg 28 0000000000000000
 * /@reg 29 0000000140023e90                    */
/************************************************/
architecture_state::
architecture_state(system_config *sc, 
		   evaluation_result *e){
  for(int i=0; i<32; i++) r[i].init(0);
  for(int i=0; i<32; i++) f[i].init(0);

  char *program_name = sc->program_name;
  FILE *fp = fp = fopen(program_name, "r");
  if(fp==NULL){
    printf("Bad file name: %s\n", 
	    program_name);
    exit(1);
  }

 /*** File Format Check ***/
  DATA_TYPE dat;
  char buf[4096];
  fgets(buf, 4096, fp);
  if(strncmp(buf+3, "SimAlpha", 8)){
    printf("%s: Not a SimAlpha data file.\n", 
	   program_name);
    exit(1);
  }
  
  while(!feof(fp)){
    int i;
    fgets(buf, 4096, fp);
    if(*buf=='/' && *(buf+1)=='@'){
      sscanf(buf+6,"%d %llx\n", &i, &dat);
      if(i==32) pc.init(dat); 
      else r[i].init(dat);
    }
    if(i==32) break;
  }
  fclose(fp);
}

uint64_t architecture_state::rr(int n){
  if(n==0) return pc.ld();
  if(n>=1  && n<=32) return r[n-1].ld();
  if(n>=33 && n<=64) return f[n-33].ld();
  printf("Error:architecture_state::rr(%d)\n",n);
  exit(1);
  return 0;
}

/************************************************/
void print_evaluation_result(struct evaluation_result *e,
			     struct system_config *ms, 
			     system_manager *sys){
  struct timeval tp;    
  struct timezone tzp;  
  gettimeofday(&tp, &tzp);
  
  uint64_t run_usec = (uint64_t)(tp.tv_sec - e->tp.tv_sec) * 1000000ull
    + (uint64_t)tp.tv_usec - (uint64_t)e->tp.tv_usec;
  fflush(stdout);
  fflush(stderr);
  int minute = (run_usec / 1000000) / 60;
  int second = (run_usec / 1000000) % 60;
  fprintf(stderr,"\n==================================");
  fprintf(stderr,"================================\n");
  fprintf(stderr,"=== %s\n", VER);
  char buf[256];
  getcwd(buf, 256);
  fprintf(stderr, "=== cwd: %s\n", buf);
  fprintf(stderr,"=== %5qd million code(%13qd code)", 
	  e->retired_inst/1000000, e->retired_inst);
  fprintf(stderr,"  %6.3f MIPS\n", 
	  (double)e->retired_inst/(double)run_usec);
  fprintf(stderr,"=== SimAlpha takes %3d min %2d second (%qd usec)\n", 
	  minute, second, run_usec);
  fprintf(stderr,"=== SimAlpha starts at %s", ctime(&e->time_begin));
  fprintf(stderr,"=== Memory %dKB used (block size %dKB, %d blocks)\n",
	  BLOCK_SIZE/1024 * e->used_memory_block,
	  BLOCK_SIZE/1024, e->used_memory_block);
  fflush(stderr);
}

/************************************************/
system_config::system_config(char *program, 
			     char **option){
  memset(this, 0, sizeof(system_config));
  strcpy(program_name, program);
  
  if(option==NULL) return;
  for(int i=0; option[i]!=NULL; i++){
    char opt[1024];
    strcpy(opt, option[i]);
    if(opt[0]!='-') continue;
    switch(opt[1]){
    case 'c': syscall_flag   = 1; break;
    case 'd': debug_flag     = 1; break;
    case 'v':
      verbose_interval =atoi(&opt[2]);
      if(verbose_interval == 0)   
	verbose_interval = 100000;
      if(opt[strlen(opt)-1]=='m') 
	verbose_interval*=1000000;
      printf("*** verbose interval: %d\n", 
	     verbose_interval);
      break;
    case 'f': 
      fd_adjust = atoi(&opt[2]);
      if(fd_adjust==0) fd_adjust=1;
      printf("*** fd_adjust=%d \n",fd_adjust);
      if(fd_adjust<0){
	printf("Error fd_adjust.\n");
	exit(1);
      }
      break;
    case 'e':
      end_tc =atoi(&opt[2]);
      if(opt[strlen(opt)-1]=='m') end_tc*=1000000;
      printf("*** Finish at %qd code execution\n",
	     end_tc);
      break;
    case 'i':
      fd_stdin_flag = 1;
      fd_stdin = open(&opt[2], O_RDONLY);
      printf("*** SimAlpha: use [%s] as STDIN.\n", 
	     &opt[2]);
      if(fd_stdin<0){
	printf("Can not open file: %s\n", &opt[2]); 
	exit(1);
      }
      break;
    default:
      printf("*** Error: [%s] Wrong option!!\n\n", 
	     opt);
      usage();
      exit(1);
    }
  }
}

/************************************************/
evaluation_result::evaluation_result(){
  memset(this,0,sizeof(class evaluation_result));

  time_begin = time(0);
  gettimeofday(&tp, &tzp);
}
/************************************************/
