/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/************************************************/
inline int nbyte(INST_TYPE ir){
  int Op = (ir>>26) & 0x3F;
  
  int n  =  (Op==0x0e || Op==0x0a) ? 1 :
    (Op==0x0d || Op==0x0c) ? 2 :
    (Op==0x22 || Op==0x26 || Op==0x28 || Op==0x2a
     || Op==0x2c || Op==0x2e) ? 4 : 8; 
  return n;
}

/************************************************/
int instruction::data_st(data_t *adr, data_t *dat){
  int Op = (ir>>26) & 0x3F;
  if(Op==0x26){ /***** STS *****/
    DATA_TYPE Rav_t;
    DATA_TYPE tmp = dat->ld();
    uint32_t int_tmp = 0;
    int_tmp = 0;
    int_tmp |= ((tmp >> 29) & 0x3FFFFFFFul); // Set 30 bit
    int_tmp |= ((tmp >> 32) & 0xc0000000ul); // Set  2 bit
    memcpy(&Rav_t, &int_tmp, 4);
    dat->st(Rav_t & 0xffffffffull);
  }

  mem->st_nbyte(nbyte(ir), adr, dat);
  return 0;
}

/************************************************/
int instruction::data_ld(data_t *adr, data_t *dat){

  mem->ld_nbyte(nbyte(ir), adr, dat); 

  int Op = (ir>>26) & 0x3F;
  if(Op==0x28 || Op==0x2a){ /**** LDL & LDL_L ****/
    DATA_TYPE Rcv_t = dat->ld();
    if(Rcv_t & BIT31) Rcv_t|=EXTND32;
    dat->st(Rcv_t);
  }
  if(Op==0x22){ /**** LDS ****/
    DATA_TYPE Rcv_t = dat->ld();
    DATA_TYPE Tmp = Rcv_t;
    Rcv_t = (Tmp >> 31) & 0x1ull;                    // Set  1 bit
    Rcv_t = (Rcv_t << 11) | (map_s(Tmp) & 0x7ffull); // Set 11 bit
    Rcv_t = (Rcv_t << 23) | (Tmp & 0x7FFFFFull);     // Set 23 bit
    Rcv_t =  Rcv_t << 29;                            // Set 29 bit
    dat->st(Rcv_t);
  }
  return 0;
}

/************************************************/
instruction::instruction(architecture_state *a, 
			 memory_system *m,
			 system_manager *s, 
			 system_config *sc,
			 evaluation_result *et){
  as  = a;
  mem = m;
  sys = s;
  e   = et;
}

/************************************************/
INST_TYPE instruction::get_ir(){
  return ir;
}

/************************************************/
int instruction::Fetch(data_t *pc){
  mem->ld_inst(pc, &ir);
  Cpc.init(pc->ld());
  Npc.init(pc->ld() + 4);
  return 0;
}

/** CMOVb case, npc and pc has the same value  **/
/************************************************/
int instruction::Fetch(data_t *pc, INST_TYPE ir_t){
  ir = ir_t;
  Cpc.init(pc->ld());
  Npc.init(pc->ld());
  return 0;
}

/** Decode IR                                  **/
/************************************************/
int instruction::Slot(){ 
  Op  = (ir>>26) &  0x3F;
  RA  = (ir>>21) &  0x1F;
  RB  = (ir>>16) &  0x1F; 
  RC  = (ir    ) &  0x1F;
  WF  = ((Op&MSK2)==0x14 || (Op&MSK2)==0x20);
  LA  = (Op==0x08 || Op==0x09);
  LD  = (Op==0x0a || Op==0x0b || Op==0x0c ||
	 (Op&MSK2)==0x20 || (Op&MSK2)==0x28);
  ST  = (Op==0x0d || Op==0x0e || Op==0x0f || 
	 (Op&MSK2)==0x24 || (Op&MSK2)==0x2c);
  BR  = ((Op&MSK4)==0x30);
  WB  = (LD || (Op&MSK2)==0x08 || Op==0x1a || 
	 Op==0x30 || Op==0x34) ? RA :
        ((Op&MSK3)==0x10 || Op==0x1c) ? RC : 31;
  Af  = (Op==0x15 || Op==0x16 || Op==0x17 ||
	 Op==0x1c ||
	 (Op&MSK2)==0x24 || (Op&MSK3)==0x30);
  Bf  = ((Op&MSK2)==0x14);
  Ai = (Op==0x08 || Op==0x09 || LD);
  Bi = (BR || (Op&MSK2)==0x10 && (ir & BIT12));

  /** For the CMOV Split Code (CMOV1) **/
  if(cmov_ir_create(ir)){ 
    CM = 1; /* This is a cmov inst. */
    RB = RC;
    Bi = 0;
  }
  return 0;
}

/************************************************/
int instruction::Rename(){ 
  return 0;
}

/** genarate the immediate data Imm            **/
/************************************************/
int instruction::Issue(){ 
  DATA_TYPE Lit, D16, D21, tmp, d21e, d16e;
  d21e = ((ir & MASK21) | EXTND21) << 2;
  d16e  = (ir & MASK16) | EXTND16;
  
  Lit = (ir>>13) & 0xFF;
  D21 = (ir & BIT20) ? d21e : (ir&MASK21)<<2;
  D16 = (ir & BIT15) ? d16e : (ir&MASK16);
  if(Op==0x09) D16 = (D16 << 16);
  
  tmp = (LA||LD||ST) ? D16 : (BR) ? D21 : Lit;
  Imm.init(tmp);
  return 0;
}

/** copy data to Rav and Rbv                   **/
/************************************************/
int instruction::RegisterRead(){
  Rav = Ai ? Imm : Af ? as->f[RA] : as->r[RA];
  Rbv = Bi ? Imm : Bf ? as->f[RB] : as->r[RB];
  return 0;
}

/** Updata Rcv, Adr and Tpc                    **/
/************************************************/
int instruction::Execute(data_t *Tpc){ 
  /*** Update Rcv ***/
  if(BR || Op==OP_JSR){
    Rcv=Npc;
  }
  else if(!LD){
    ALU(ir, &Rav, &Rbv, &Rcv); 
  }

  /*** Update Adr ***/
  Adr.init(0);
  if(LD || ST){
    ALU(ir, &Imm, &Rbv, &Adr);
  }
  
  /*** Update Npc ***/
  if(Op==OP_JSR){
    Npc = Rbv;
    Npc.st(Npc.ld() & ~3ull);
  }
  if(BR){
    BRU(ir, &Rav, &Rbv, &Npc, &Npc);
  }
  *Tpc = Npc;

  return 0;
}

/**  store Rav to mem or load to Rcv           **/
/************************************************/
int instruction::Memory(){ 
  if(ST) data_st(&Adr, &Rav);
  if(LD) data_ld(&Adr, &Rcv);
  return 0;
}

/** PAL or copy Rcv into the regfile           **/
/************************************************/
int instruction::WriteBack(){ 
  if(Op==OP_PAL){
    sys->execute_pal(this);
  }

  if(!WF && WB!=31) as->r[WB] = Rcv;
  if( WF && WB!=31) as->f[WB] = Rcv;
  return 0;
}
/************************************************/
