/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/************************************************/
int system_manager::translate_fd(DATA_TYPE fd){
  for(int i=0; i<FD_MAPPING_MAX; i++){ 
    if(fd_table[i] == fd) return i;
  }
  printf("*** Error in translate_fd\n");
  exit(1);
  return 0;
}

/************************************************/
system_manager::
system_manager(architecture_state *as_t, 
	       memory_system *mem_t,
	       struct system_config *ms_t, 
	       struct evaluation_result *e_t){
  as   =  as_t;
  mem  = mem_t;
  ms   =  ms_t;
  e    =   e_t;
  uniq    = 0;
  running = 1;
  
  /** for example, fd_table[6]=7        **/
  /** maps real fd(6) to logical fd(7). **/
  /***************************************/
  fd_table[0] = 0; // STDIN
  fd_table[1] = 1; // STDOUT
  fd_table[2] = 2; // STDERR
  for(int i=3; i<FD_MAPPING_MAX; i++)
    fd_table[i] = i + 2 - ms->fd_adjust;
}

/************************************************/
void system_manager::execute_pal(instruction *p){
  INST_TYPE ir = p->get_ir();
  int64_t ret=0;
  int F = ms->syscall_flag;
  data_t *r = as->r;

  if((ir & 0xFFFF)!=0x0083){
    switch((ir & 0xFFFF)){
    case 0x009e: // rdunique(Read Unique Value)
      r[0].init(uniq);
      if(F) printf("%8qd: call_pal rduniq = %llx\n", 
		   e->retired_inst, uniq);
      return;
    case 0x009f: // wrunique(Write Unique Value)
      uniq  = r[16].ld();
      if(F) printf("%8qd: call_pal wrruniq = %llx\n", 
		   e->retired_inst, r[16].ld());
      return;
    default:
      if(F){
	printf("*** PAL (ir & 0xFFFF)!=0x0083  Function %x", 
	       (ir & 0xffff));
	printf("  TC:%qd default. %llx\n",
	       e->retired_inst, r[0].ld());
      }
      return; 
    }
  }

  if(r[0].ld()==0){
    r[19].init(0);
    return;
  }


  switch(r[0].ld()){
    /*****************************************/
  case ASYS_UNIMPLEMENT:
    {
      ret = -1;
      r[0].init(0);
      r[19].init(ret);
      if(F) printf("%8qd: SYS_226 = %lld\n",
		   e->retired_inst, ret);
      break;
    }
    /*****************************************/
  case ASYS_unlink:
    {
      char buf[512]; /** file name **/
      for(int i=0; i<=512; i++){
	data_t d;
	data_t adr;
	adr.init(r[16].ld()+i);
	mem->ld_nbyte(1, &adr, &d);
	buf[i] = d.ld();
	if(buf[i]=='\0') break;
      }
      if(ms->fd_adjust){ /** rename **/
	buf[1]='S';
	buf[2]='A';
	buf[3]='0'+ ms->fd_adjust;
      }

      ret = unlink(buf);
      if(ret==-1){
	r[0].init(errno);
	r[19].init(ret);
      }
      else{
	r[19].init(0);
      }

      if(F) printf("%8qd: unlink(%s) = %qd\n",
		   e->retired_inst, buf, ret);
      break;
    }
    /*****************************************/
  case ASYS_open:{
    int fd_t;
    char buf[512]; /** file name **/
    for(int i=0; i<=512; i++){
	data_t d;
	data_t adr;
	adr.init(r[16].ld()+i);
	mem->ld_nbyte(1, &adr, &d);
	buf[i] = d.ld();
	if(buf[i]=='\0') break;
    }
    /** RDONLY or RDWR **/
    if(r[17].ld()==0 || r[17].ld()==2){
      fd_t = open(buf, r[17].ld(), r[18].ld());
    }
    else{
      if(ms->fd_adjust){ /** rename **/
	buf[1]='S';
	buf[2]='A';
	buf[3]='0'+ ms->fd_adjust;
      }
      int flag = r[17].ld() | O_CREAT;
      fd_t = open(buf, flag, r[18].ld());
    }
    
    /* mapping from the real to the logical */
    if(fd_t==-1){
      ret = -1;
      r[0].init(errno);
      r[19].init(~0ull);
    }
    else{
      ret = fd_table[fd_t];
      r[0].init(ret);
      r[19].init(0);
    }
    
    if(F){
      printf("%8qd: open(\"%s\", 0x%qx, 0x%qx)",
	     e->retired_inst, 
	     buf, r[17].ld(), r[18].ld());
      printf("= logic %qd(real %d)\n",
	     ret, fd_t);
    }
    break;
  }
    /*****************************************/
    /*** Note that sigreturn updates as->pc ***/
  case ASYS_sigreturn:{
    uint64_t pt[OSF_SIGCONTEXT_SIZE/8];
    char *buf = (char *)pt;
    
    data_t d;
    data_t adr;
    for(int i=0; i<OSF_SIGCONTEXT_SIZE; i++){
      adr.init(r[16].ld() + i);
      mem->ld_nbyte(1, &adr, &d);
      buf[i] = d.ld();
    }
    for(int i=0; i<31; i++) r[i].init(pt[i+4]);
    for(int i=0; i<31; i++) as->f[i].init(pt[i+37]);
    as->pc.init(pt[2]);  /** Update PC **/
    
    if(F) printf("%8qd: sigreturn(0x%qx), next pc:0x%qx\n", 
		 e->retired_inst, r[16].ld(), as->pc.ld());
    break;
  }
    /*****************************************/
  case ASYS_close:{
    int fd_t = translate_fd(r[16].ld());
    
    if(fd_t!=0 && fd_t!=1 && fd_t!=2){
      ret = close(fd_t);
      r[0].init(ret);
    }
    r[19].init(0);
    if(F) printf("%8qd: close(%qd(%d)) = %qd\n", 
		 e->retired_inst, r[16].ld(),
		 fd_t, ret);
    break;
  }
    /*****************************************/
  case ASYS_getpagesize:{
    ret = 8192; /* getpagesize(); */
    r[0].init(ret);
    r[19].init(0);
    if(F) printf("%8qd: getpagesize() = 0x%llx\n", 
		 e->retired_inst, ret);
    break;
  }
    /*****************************************/
  case ASYS_sigprocmask:
    ret = 0;
    r[0].init(ret);
    r[19].init(ret);
    if(F) printf("%8qd: sigprocmask(%qd,%qx) = %qd\n", 
		 e->retired_inst, r[16].ld(), 
		 r[17].ld(), ret);
    break;
    /*****************************************/
  case ASYS_getrusage: // Not Checked 
    {
      struct rusage ru;
      ret = getrusage((enum __rusage_who)r[16].ld(), &ru);

      for(int i=0; i<OSF_RUSAGE_SIZE; i++){
	data_t adr; adr.init(r[17].ld()+i);
	data_t dat; dat.init(0);
	mem->st_nbyte(1, &adr, &dat);
      }
      
      r[0].init(ret);
      r[19].init(0);

      if(F) printf("%8qd: getrusage(%qd, %qx) = %qd\n", 
		   e->retired_inst, 
		   r[16].ld(), r[17].ld(), ret);
      break;
    }
    /*****************************************/
  case 	ASYS_gettimeofday:
    {
      r[19].init(0);
      if(F) printf("%8qd: gettimeofday(%qd,%qx) = %qd\n", 
		   e->retired_inst, r[16].ld(), 
		   r[17].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_fcntl:
    {
      ret = 0;
      r[0].init(ret);
      r[19].init(0);
      if(F) printf("%8qd: fcntl(%qd,%qx) = %qd\n", 
		   e->retired_inst, r[16].ld(), 
		   r[17].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_sigaction:
    {
      ret = 0;
      r[0].init(ret);
      r[18].init(r[17].ld());
      r[19].init(0);
      if(F) printf("%8qd: sigaction(%qx,%qx,%qx,%qx) = %qd\n", 
		   e->retired_inst,
		   r[16].ld(), r[17].ld(),
		   r[18].ld(), r[19].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_osf_getsysinfo:
    {
      ret = 0;
      r[0].init(ret);
      r[19].init(0);
      if(F) printf("%8qd: osf_getsysinfo(%qx,%qx,%qx,%qx) = %qd\n", 
		   e->retired_inst, 
		   r[16].ld(), r[17].ld(),
		   r[18].ld(), r[19].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_osf_setsysinfo:
    {
      ret = 0;
      r[0].init(ret);
      //    r[20]= ret;
      r[19].init(0);
      if(F) printf("%8qd: osf_setsysinfo(%qx,%qx,%qx,%qx) = %qd\n", 
		   e->retired_inst, 
		   r[16].ld(), r[17].ld(),
		   r[18].ld(), r[19].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_getuid:
    {
      ret = getuid();
      r[0].init(ret);
      r[19].init(0);
      r[20].init(geteuid());
      if(F) printf("%8qd: getuid() = %qd\n", e->retired_inst, ret);
      break;
    }
    /*****************************************/
  case ASYS_getgid:
    {
      ret = getgid();
      r[0].init(ret);
      r[19].init(0);
      r[20].init(getegid());
      if(F) printf("%8qd: getgid(0x%qx) = %qd\n", e->retired_inst, 
		   r[16].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_brk:
    {
      ret = r[16].ld();
      r[0].init(ret);
      r[19].init(0);
      if(F) printf("%8qd: brk(0x%qx) = %qx\n", e->retired_inst, 
		   r[16].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_getpid:
    {
      ret = getpid();
      r[0].init(ret);
      r[19].init(0);
      if(F) printf("%8qd: getpid() = %qx\n", 
		   e->retired_inst, ret);
      break;
    }
    /*****************************************/
    /*
  case ASYS_stat: 
    {
      if(F) printf("%8qd: stat() = %qx\n", 
		   e->retired_inst, ret);
      break;
    }
    */
    /*****************************************/
  case ASYS_fstat:
    {
      int fd_t = translate_fd(r[16].ld());

      struct stat sbuf;
      ret = fstat(fd_t,  &sbuf);

      struct osf_statbuf asbuf;
      memset(&asbuf, 0, sizeof(struct osf_statbuf));
      asbuf.st_dev     = sbuf.st_dev;
      asbuf.st_ino     = sbuf.st_ino;
      asbuf.st_mode    = sbuf.st_mode;
      asbuf.st_nlink   = sbuf.st_nlink;
      asbuf.st_uid     = sbuf.st_uid;
      asbuf.st_gid     = sbuf.st_gid;
      asbuf.st_rdev    = sbuf.st_rdev;
      asbuf.st_size    = sbuf.st_size;
      asbuf.st_atime   = sbuf.st_atime;
      asbuf.st_mtime   = sbuf.st_mtime;
      asbuf.st_ctime   = sbuf.st_ctime;
      asbuf.st_blksize = sbuf.st_blksize;
      asbuf.st_blocks  = sbuf.st_blocks;

      char *buf = (char *)&asbuf;

      for(int i=0; i<OSF_FSTAT_SIZE; i++){
	data_t adr; 
	adr.init(r[17].ld()+i);
	data_t dat; 
	dat.init(buf[i]);
	mem->st_nbyte(1, &adr, &dat);
      }
      
      r[0].init(ret);
      r[19].init(0);
      if(F) printf("%8qd: fstat(logic %qd(real %d), 0x%qx) = %qd\n", 
		   e->retired_inst, r[16].ld(), 
		   fd_t, r[17].ld(), ret);
      break; 
    }
    /*****************************************/
  case ASYS_ioctl: /** now write 0, should modify **/
    {
      int fd_t = translate_fd(r[16].ld());

      switch(r[17].ld()){
      case 0x40067408:      // OSF_TIOCGRETP 
	{
	  struct osf_sgttyb {
	    uint8_t  sg_ispeed;     /* input speed */
	    uint8_t  sg_ospeed;     /* output speed */
	    uint8_t  sg_erase;      /* erase character */
	    uint8_t  sg_kill;       /* kill character */
	    uint16_t sg_flags;      /* mode flags */
	  };
	  
	  struct termios lbuf;
	  ret = tcgetattr(fd_t, &lbuf);

	  /*** store 6 byte, value of 0 ***/
	  for(int i=0; i<6; i++){
	    data_t adr; adr.init(r[18].ld()+i);
	    data_t dat; dat.init(0);
	    mem->st_nbyte(1, &adr, &dat);
	  }

	  if(ret!=-1){
	    r[0].init(ret);
	    r[19].init(0); // A3
	  }
	  else{
	    r[0].init(errno);
	    r[19].init(ret);
	  }
	  break;
	}

	
      default:
	ret = 0;
	r[19].init(ret);
	if(F) printf("*** ioctl default\n");
	break;
      }

      if(F)
	printf("%8qd: ioctl(logic %qd(real %d), 0x%qx, 0x%qx) = 0x%qx\n", 
	       e->retired_inst, r[16].ld(), fd_t,
	       r[17].ld(), r[18].ld(), ret); 
      break;
    }

    /*****************************************/
  case ASYS_read:
    {
      char *buf = (char *)calloc(1, r[18].ld());
      if(buf==NULL) printf("*** malloc error in syscall.cc::read\n");

      int fd_t = translate_fd(r[16].ld());
      if(ms->fd_stdin_flag && fd_t==0) fd_t=ms->fd_stdin; /* STDIN */
      
      ret = read(fd_t, buf, r[18].ld());
      
      /*      for(int i=0; i<ret; i++){  */  // This is better.
      for(int i=0; i<(int)r[18].ld(); i++){ // simple scalar imple!
	DATA_TYPE d = buf[i];
	data_t adr; adr.init(r[17].ld()+i);
	data_t dat; dat.init(d);
	mem->st_nbyte(1, &adr, &dat);
      }
      r[0].init(ret);
      r[19].init(0);
      free(buf);

      if(F) printf("%8qd: read(logic %qd(real %d),%qx,0x%qx) = 0x%qx\n", 
		   e->retired_inst, r[16].ld(),
		   fd_t, r[17].ld(), r[18].ld(), 
		   ret);
      break;
    }
    
    /*****************************************/
  case ASYS_write:
    {
      char *buf = (char *)malloc(r[18].ld());
      if(buf==NULL) printf("*** malloc error in syscall.cc::write\n");
	
      for(int i=0; i<(int)r[18].ld(); i++){
	data_t d;
	data_t adr;
	adr.init(r[17].ld()+i);
	mem->ld_nbyte(1, &adr, &d);
	buf[i] = d.ld();
      }

      int fd_t = translate_fd(r[16].ld());

      ret = write(fd_t, (void *)buf, (long)r[18].ld());
      r[0].init(ret);
      r[19].init(0);
      free(buf);

      if(F) printf("%8qd: write(%qd(%d), 0x%qx, %qd) = %qd\n", 
		   e->retired_inst, r[16].ld(),
		   fd_t, r[17].ld(), r[18].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_lseek:
    {
      int fd_t = translate_fd(r[16].ld());
      if(ms->fd_stdin_flag && fd_t==0) fd_t=ms->fd_stdin; /* STDIN */

      ret = lseek(fd_t, r[17].ld(), r[18].ld());
      if(ret==-1){
	r[0].init(errno);
	r[19].init(~0);
      }
      else{
	r[0].init(ret);
	r[19].init(0);
      }
      if(F) printf("%8qd: lseek(logic %qd(real %d), %qd, %qd) = 0x%qx\n",
		   e->retired_inst, 
		   r[16].ld(), fd_t, 
		   r[17].ld(), r[18].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_getrlimit:
    {
      ret = 0;
      r[19].init(ret);
      if(F) printf("%8qd: getrlimit(%qd,%qd,%qd) = %qd\n",
		   e->retired_inst, r[16].ld(), 
		   r[17].ld(), r[18].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_setrlimit:
    {
      ret = 0;
      r[19].init(ret);
      if(F) printf("%8qd: setrlimit(%qd,%qd,%qd) = %qd\n",
		   e->retired_inst, r[16].ld(), 
		   r[17].ld(), r[18].ld(), ret);
      break;
    }
    /*****************************************/
  case ASYS_exit:
    {
      ret = 0;
      running = 0;
      if(F) printf("%8qd: exit() = %qd\n", e->retired_inst, ret);
      break;
    }    

  default:
    {
      printf("TC: %qd systemcall not implemented!!: %qd\n" ,
	     e->retired_inst, r[0].ld());
      r[0].init(0);
      r[19].init((uint64_t)-1);
    }
  }
  fflush(stdout);
}
/************************************************/
